// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

import baseConfig from "../rollup-base"

export default baseConfig(__dirname, {
    "@microsoft/signalr": "signalR",
});
