// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

package com.microsoft.signalr;

interface WebSocketOnClosedCallback {
    void invoke(Integer code, String reason);
}
