// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.HttpSys.Internal;

internal enum SslStatus : byte
{
    Insecure,
    NoClientCert,
    ClientCert
}
