// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.DotNet.ApiCompat
{
    public class Program
    {
        public static int Main(string[] args)
        {
            return new ApiCompatRunner().Run(args);   
        }
    }
}
