// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.Authentication.Negotiate;

/// <summary>
/// Default values used by Negotiate authentication.
/// </summary>
public static class NegotiateDefaults
{
    /// <summary>
    /// Default value for AuthenticationScheme used to identify the Negotiate auth handler.
    /// </summary>
    public const string AuthenticationScheme = "Negotiate";
}
