// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.Server.Kestrel.Core.Internal.Http3;

[Flags]
internal enum StreamCompletionFlags
{
    None = 0,
    EndStreamReceived = 1,
    AbortedRead = 2,
    Aborted = 4,
    Completed = 8,
}
