// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.Server.HttpSys;

internal interface IRequestContextFactory
{
    RequestContext CreateRequestContext(uint? bufferSize, ulong requestId);
}
