// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.Authentication.OAuth;

/// <summary>
/// Defaults for OAuth authentication.
/// </summary>
public static class OAuthDefaults
{
    /// <summary>
    /// The default display name for OAuth authentication.
    /// </summary>
    public static readonly string DisplayName = "OAuth";
}
