// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.Authentication.Twitter;

internal sealed class TwitterError
{
    public int Code { get; set; }
    public string? Message { get; set; }
}
