// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.AspNetCore.Server.HttpSys;

/// <summary>
/// Constants for HttpSys.
/// </summary>
public static class HttpSysDefaults
{
    /// <summary>
    /// The name of the authentication scheme used.
    /// </summary>
    public const string AuthenticationScheme = "Windows";
}
