// <auto-generated/>
#pragma warning disable 1591
namespace Microsoft.AspNetCore.Hosting.Views
{
    #line hidden
    using System.Threading.Tasks;
#nullable restore
#line 1 "Views/ErrorPage.cshtml"
using System;

#line default
#line hidden
#nullable disable
#nullable restore
#line 2 "Views/ErrorPage.cshtml"
using System.Globalization;

#line default
#line hidden
#nullable disable
#nullable restore
#line 3 "Views/ErrorPage.cshtml"
using System.Linq;

#line default
#line hidden
#nullable disable
#nullable restore
#line 4 "Views/ErrorPage.cshtml"
using System.Net;

#line default
#line hidden
#nullable disable
#nullable restore
#line 5 "Views/ErrorPage.cshtml"
using System.Reflection;

#line default
#line hidden
#nullable disable
#nullable restore
#line 6 "Views/ErrorPage.cshtml"
using Microsoft.AspNetCore.Hosting.Views;

#line default
#line hidden
#nullable disable
    internal sealed class ErrorPage : Microsoft.Extensions.RazorViews.BaseView
    {
        #pragma warning disable 1998
        public async override global::System.Threading.Tasks.Task ExecuteAsync()
        {
            WriteLiteral("\r\n");
#nullable restore
#line 17 "Views/ErrorPage.cshtml"
  
    var location = string.Empty;

#line default
#line hidden
#nullable disable
            WriteLiteral("<!DOCTYPE html>\r\n<html");
            BeginWriteAttribute("lang", " lang=\"", 366, "\"", 427, 1);
#nullable restore
#line 21 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 373, CultureInfo.CurrentUICulture.TwoLetterISOLanguageName, 373, 54, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(@" xmlns=""http://www.w3.org/1999/xhtml"">
    <head>
        <meta charset=""utf-8"" />
        <title>Internal Server Error</title>
        <style>
            body {
    font-family: 'Segoe UI', Tahoma, Arial, Helvetica, sans-serif;
    font-size: .813em;
    color: #222;
}

h1, h2, h3, h4, h5 {
    /*font-family: 'Segoe UI',Tahoma,Arial,Helvetica,sans-serif;*/
    font-weight: 100;
}

h1 {
    color: #44525e;
    margin: 15px 0 15px 0;
}

h2 {
    margin: 10px 5px 0 0;
}

h3 {
    color: #363636;
    margin: 5px 5px 0 0;
}

code {
    font-family: Consolas, ""Courier New"", courier, monospace;
}

body .titleerror {
    padding: 3px 3px 6px 3px;
    display: block;
    font-size: 1.5em;
    font-weight: 100;
}

body .location {
    margin: 3px 0 10px 30px;
}

#header {
    font-size: 18px;
    padding: 15px 0;
    border-top: 1px #ddd solid;
    border-bottom: 1px #ddd solid;
    margin-bottom: 0;
}

    #header li {
        display: inline;
        margin: 5p");
            WriteLiteral(@"x;
        padding: 5px;
        color: #a0a0a0;
        cursor: pointer;
    }

    #header .selected {
        background: #44c5f2;
        color: #fff;
    }

#stackpage ul {
    list-style: none;
    padding-left: 0;
    margin: 0;
    /*border-bottom: 1px #ddd solid;*/
}

#stackpage .details {
    font-size: 1.2em;
    padding: 3px;
    color: #000;
}

#stackpage .stackerror {
    padding: 5px;
    border-bottom: 1px #ddd solid;
}


#stackpage .frame {
    padding: 0;
    margin: 0 0 0 30px;
}

    #stackpage .frame h3 {
        padding: 2px;
        margin: 0;
    }

#stackpage .source {
    padding: 0 0 0 30px;
}

    #stackpage .source ol li {
        font-family: Consolas, ""Courier New"", courier, monospace;
        white-space: pre;
        background-color: #fbfbfb;
    }

#stackpage .frame .source .highlight {
    border-left: 3px solid red;
    margin-left: -3px;
    font-weight: bold;
}

#stackpage .frame .source .highlight li span {
    co");
            WriteLiteral(@"lor: #FF0000;
}

#stackpage .source ol.collapsible li {
    color: #888;
}

    #stackpage .source ol.collapsible li span {
        color: #606060;
    }

.page table {
    border-collapse: separate;
    border-spacing: 0;
    margin: 0 0 20px;
}

.page th {
    vertical-align: bottom;
    padding: 10px 5px 5px 5px;
    font-weight: 400;
    color: #a0a0a0;
    text-align: left;
}

.page td {
    padding: 3px 10px;
}

.page th, .page td {
    border-right: 1px #ddd solid;
    border-bottom: 1px #ddd solid;
    border-left: 1px transparent solid;
    border-top: 1px transparent solid;
    box-sizing: border-box;
}

    .page th:last-child, .page td:last-child {
        border-right: 1px transparent solid;
    }

.page .length {
    text-align: right;
}

a {
    color: #1ba1e2;
    text-decoration: none;
}

    a:hover {
        color: #13709e;
        text-decoration: underline;
    }

.showRawException {
    cursor: pointer;
    color: #44c5f2;
    bac");
            WriteLiteral(@"kground-color: transparent;
    font-size: 1.2em;
    text-align: left;
    text-decoration: none;
    display: inline-block;
    border: 0;
    padding: 0;
}

.rawExceptionStackTrace {
    font-size: 1.2em;
}

.rawExceptionBlock {
    border-top: 1px #ddd solid;
    border-bottom: 1px #ddd solid;
}

.showRawExceptionContainer {
    margin-top: 10px;
    margin-bottom: 10px;
}

.expandCollapseButton {
    cursor: pointer;
    float: left;
    height: 16px;
    width: 16px;
    font-size: 10px;
    position: absolute;
    left: 10px;
    background-color: #eee;
    padding: 0;
    border: 0;
    margin: 0;
}

        </style>
    </head>
    <body>
        <h1>An error occurred while starting the application.</h1>
");
#nullable restore
#line 232 "Views/ErrorPage.cshtml"
         foreach (var errorDetail in Model.ErrorDetails)
        {

#line default
#line hidden
#nullable disable
            WriteLiteral("            <div class=\"titleerror\">");
#nullable restore
#line 234 "Views/ErrorPage.cshtml"
                               Write(errorDetail.Error!.GetType().Name);

#line default
#line hidden
#nullable disable
            WriteLiteral(": ");
#nullable restore
#line 234 "Views/ErrorPage.cshtml"
                                                                           Output.Write(HtmlEncodeAndReplaceLineBreaks(errorDetail.Error!.Message)); 

#line default
#line hidden
#nullable disable
            WriteLiteral("</div>\r\n");
#nullable restore
#line 235 "Views/ErrorPage.cshtml"

            var firstFrame = errorDetail.StackFrames.FirstOrDefault();
            if (firstFrame != null)
            {
                location = firstFrame.Function;
            }
            if (!string.IsNullOrEmpty(location) && firstFrame != null && !string.IsNullOrEmpty(firstFrame.File))
            {

#line default
#line hidden
#nullable disable
            WriteLiteral("                <p class=\"location\">");
#nullable restore
#line 243 "Views/ErrorPage.cshtml"
                               Write(location);

#line default
#line hidden
#nullable disable
            WriteLiteral(" in <code");
            BeginWriteAttribute("title", " title=\"", 4864, "\"", 4888, 1);
#nullable restore
#line 243 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 4872, firstFrame.File, 4872, 16, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(">");
#nullable restore
#line 243 "Views/ErrorPage.cshtml"
                                                                           Write(System.IO.Path.GetFileName(firstFrame.File));

#line default
#line hidden
#nullable disable
            WriteLiteral("</code>, line ");
#nullable restore
#line 243 "Views/ErrorPage.cshtml"
                                                                                                                                     Write(firstFrame.Line);

#line default
#line hidden
#nullable disable
            WriteLiteral("</p>\r\n");
#nullable restore
#line 244 "Views/ErrorPage.cshtml"
            }
            else if (!string.IsNullOrEmpty(location))
            {

#line default
#line hidden
#nullable disable
            WriteLiteral("                <p class=\"location\">");
#nullable restore
#line 247 "Views/ErrorPage.cshtml"
                               Write(location);

#line default
#line hidden
#nullable disable
            WriteLiteral("</p>\r\n");
#nullable restore
#line 248 "Views/ErrorPage.cshtml"
            }
            else
            {

#line default
#line hidden
#nullable disable
            WriteLiteral("                <p class=\"location\">Unknown location</p>\r\n");
#nullable restore
#line 252 "Views/ErrorPage.cshtml"
            }

            var reflectionTypeLoadException = errorDetail.Error as ReflectionTypeLoadException;
            if (reflectionTypeLoadException != null)
            {
                if (reflectionTypeLoadException.LoaderExceptions.Length > 0)
                {

#line default
#line hidden
#nullable disable
            WriteLiteral("                    <h3>Loader Exceptions:</h3>\r\n                    <ul>\r\n");
#nullable restore
#line 261 "Views/ErrorPage.cshtml"
                         foreach (var ex in reflectionTypeLoadException.LoaderExceptions)
                        {
                            if (ex != null)
                            {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                <li>");
#nullable restore
#line 265 "Views/ErrorPage.cshtml"
                               Write(ex.Message);

#line default
#line hidden
#nullable disable
            WriteLiteral("</li>\r\n");
#nullable restore
#line 266 "Views/ErrorPage.cshtml"
                            }
                        }

#line default
#line hidden
#nullable disable
            WriteLiteral("                    </ul>\r\n");
#nullable restore
#line 269 "Views/ErrorPage.cshtml"
                }
            }
        }

#line default
#line hidden
#nullable disable
            WriteLiteral("        <div id=\"stackpage\" class=\"page\">\r\n            <ul>\r\n");
#nullable restore
#line 274 "Views/ErrorPage.cshtml"
                  
                    var exceptionCount = 0;
                    var stackFrameCount = 0;
                    var exceptionDetailId = "";
                    var frameId = "";
                

#line default
#line hidden
#nullable disable
#nullable restore
#line 280 "Views/ErrorPage.cshtml"
                 foreach (var errorDetail in Model.ErrorDetails)
                {
                    exceptionCount++;
                    exceptionDetailId = "exceptionDetail" + exceptionCount;

#line default
#line hidden
#nullable disable
            WriteLiteral("                    <li>\r\n                        <h2 class=\"stackerror\">");
#nullable restore
#line 285 "Views/ErrorPage.cshtml"
                                          Write(errorDetail.Error!.GetType().Name);

#line default
#line hidden
#nullable disable
            WriteLiteral(": ");
#nullable restore
#line 285 "Views/ErrorPage.cshtml"
                                                                              Write(errorDetail.Error!.Message);

#line default
#line hidden
#nullable disable
            WriteLiteral("</h2>\r\n                        <ul>\r\n");
#nullable restore
#line 287 "Views/ErrorPage.cshtml"
                         foreach (var frame in errorDetail.StackFrames)
                        {
                            stackFrameCount++;
                            frameId = "frame" + stackFrameCount;

#line default
#line hidden
#nullable disable
            WriteLiteral("                            <li class=\"frame\"");
            BeginWriteAttribute("id", " id=\"", 6857, "\"", 6870, 1);
#nullable restore
#line 291 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 6862, frameId, 6862, 8, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(">\r\n");
#nullable restore
#line 292 "Views/ErrorPage.cshtml"
                                 if (string.IsNullOrEmpty(frame.File))
                                {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                    <h3>");
#nullable restore
#line 294 "Views/ErrorPage.cshtml"
                                   Write(frame.Function);

#line default
#line hidden
#nullable disable
            WriteLiteral("</h3>\r\n");
#nullable restore
#line 295 "Views/ErrorPage.cshtml"
                                }
                                else
                                {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                    <h3>");
#nullable restore
#line 298 "Views/ErrorPage.cshtml"
                                   Write(frame.Function);

#line default
#line hidden
#nullable disable
            WriteLiteral(" in <code");
            BeginWriteAttribute("title", " title=\"", 7215, "\"", 7234, 1);
#nullable restore
#line 298 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 7223, frame.File, 7223, 11, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(">");
#nullable restore
#line 298 "Views/ErrorPage.cshtml"
                                                                                Write(System.IO.Path.GetFileName(frame.File));

#line default
#line hidden
#nullable disable
            WriteLiteral("</code></h3>\r\n");
#nullable restore
#line 299 "Views/ErrorPage.cshtml"
                                }

#line default
#line hidden
#nullable disable
            WriteLiteral("\r\n");
#nullable restore
#line 301 "Views/ErrorPage.cshtml"
                                 if (frame.Line != 0 && frame.ContextCode.Any())
                                {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                    <button class=\"expandCollapseButton\" data-frameId=\"");
#nullable restore
#line 303 "Views/ErrorPage.cshtml"
                                                                                  Write(frameId);

#line default
#line hidden
#nullable disable
            WriteLiteral("\">+</button>\r\n                                    <div class=\"source\">\r\n");
#nullable restore
#line 305 "Views/ErrorPage.cshtml"
                                         if (frame.PreContextCode.Any())
                                        {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                            <ol");
            BeginWriteAttribute("start", " start=\"", 7774, "\"", 7803, 1);
#nullable restore
#line 307 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 7782, frame.PreContextLine, 7782, 21, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(" class=\"collapsible\">\r\n");
#nullable restore
#line 308 "Views/ErrorPage.cshtml"
                                                 foreach (var line in frame.PreContextCode)
                                                {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                                    <li><span>");
#nullable restore
#line 310 "Views/ErrorPage.cshtml"
                                                         Write(line);

#line default
#line hidden
#nullable disable
            WriteLiteral("</span></li>\r\n");
#nullable restore
#line 311 "Views/ErrorPage.cshtml"
                                                }

#line default
#line hidden
#nullable disable
            WriteLiteral("                                            </ol>\r\n");
#nullable restore
#line 313 "Views/ErrorPage.cshtml"
                                        }

#line default
#line hidden
#nullable disable
            WriteLiteral("\r\n                                        <ol");
            BeginWriteAttribute("start", " start=\"", 8242, "\"", 8261, 1);
#nullable restore
#line 315 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 8250, frame.Line, 8250, 11, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(" class=\"highlight\">\r\n");
#nullable restore
#line 316 "Views/ErrorPage.cshtml"
                                             foreach (var line in frame.ContextCode)
                                            {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                                <li><span>");
#nullable restore
#line 318 "Views/ErrorPage.cshtml"
                                                     Write(line);

#line default
#line hidden
#nullable disable
            WriteLiteral("</span></li>\r\n");
#nullable restore
#line 319 "Views/ErrorPage.cshtml"
                                            }

#line default
#line hidden
#nullable disable
            WriteLiteral("                                        </ol>\r\n\r\n");
#nullable restore
#line 322 "Views/ErrorPage.cshtml"
                                         if (frame.PostContextCode.Any())
                                        {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                            <ol");
            BeginWriteAttribute("start", " start=\'", 8754, "\'", 8779, 1);
#nullable restore
#line 324 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 8762, frame.Line + 1, 8762, 17, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(" class=\"collapsible\">\r\n");
#nullable restore
#line 325 "Views/ErrorPage.cshtml"
                                                 foreach (var line in frame.PostContextCode)
                                                {

#line default
#line hidden
#nullable disable
            WriteLiteral("                                                    <li><span>");
#nullable restore
#line 327 "Views/ErrorPage.cshtml"
                                                         Write(line);

#line default
#line hidden
#nullable disable
            WriteLiteral("</span></li>\r\n");
#nullable restore
#line 328 "Views/ErrorPage.cshtml"
                                                }

#line default
#line hidden
#nullable disable
            WriteLiteral("                                            </ol>\r\n");
#nullable restore
#line 330 "Views/ErrorPage.cshtml"
                                        }

#line default
#line hidden
#nullable disable
            WriteLiteral("                                    </div>\r\n");
#nullable restore
#line 332 "Views/ErrorPage.cshtml"
                                }

#line default
#line hidden
#nullable disable
            WriteLiteral("                            </li>\r\n");
#nullable restore
#line 334 "Views/ErrorPage.cshtml"
                        }

#line default
#line hidden
#nullable disable
            WriteLiteral(@"                        </ul>
                    </li>
                    <li>
                        <br/>
                        <div class=""rawExceptionBlock"">
                            <div class=""showRawExceptionContainer"">
                                <button class=""showRawException"" data-exceptionDetailId=""");
#nullable restore
#line 341 "Views/ErrorPage.cshtml"
                                                                                    Write(exceptionDetailId);

#line default
#line hidden
#nullable disable
            WriteLiteral("\">Show raw exception details</button>\r\n                            </div>\r\n                            <div");
            BeginWriteAttribute("id", " id=\"", 9770, "\"", 9793, 1);
#nullable restore
#line 343 "Views/ErrorPage.cshtml"
WriteAttributeValue("", 9775, exceptionDetailId, 9775, 18, false);

#line default
#line hidden
#nullable disable
            EndWriteAttribute();
            WriteLiteral(" class=\"rawExceptionDetails\">\r\n                                <pre class=\"rawExceptionStackTrace\">");
#nullable restore
#line 344 "Views/ErrorPage.cshtml"
                                                               Write(errorDetail.Error!.ToString());

#line default
#line hidden
#nullable disable
            WriteLiteral("</pre>\r\n                            </div>\r\n                        </div>\r\n                    </li>\r\n");
#nullable restore
#line 348 "Views/ErrorPage.cshtml"
                }

#line default
#line hidden
#nullable disable
            WriteLiteral("            </ul>\r\n        </div>\r\n");
#nullable restore
#line 351 "Views/ErrorPage.cshtml"
         if (Model.ShowRuntimeDetails) {

#line default
#line hidden
#nullable disable
            WriteLiteral("            <footer>\r\n                ");
#nullable restore
#line 353 "Views/ErrorPage.cshtml"
           Write(Model.RuntimeDisplayName);

#line default
#line hidden
#nullable disable
            WriteLiteral(" ");
#nullable restore
#line 353 "Views/ErrorPage.cshtml"
                                     Write(Model.RuntimeArchitecture);

#line default
#line hidden
#nullable disable
            WriteLiteral(" v");
#nullable restore
#line 353 "Views/ErrorPage.cshtml"
                                                                  Write(Model.ClrVersion);

#line default
#line hidden
#nullable disable
            WriteLiteral(" &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;Microsoft.AspNetCore.Hosting version ");
#nullable restore
#line 353 "Views/ErrorPage.cshtml"
                                                                                                                                                               Write(Model.CurrentAssemblyVesion);

#line default
#line hidden
#nullable disable
            WriteLiteral(" &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp; ");
#nullable restore
#line 353 "Views/ErrorPage.cshtml"
                                                                                                                                                                                                                                  Write(Model.OperatingSystemDescription);

#line default
#line hidden
#nullable disable
            WriteLiteral(" &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;<a href=\"http://go.microsoft.com/fwlink/?LinkId=517394\">Need help?</a>\r\n            </footer>\r\n");
#nullable restore
#line 355 "Views/ErrorPage.cshtml"
        }

#line default
#line hidden
#nullable disable
            WriteLiteral(@"        <script>
            //<!--
            (function (window, undefined) {
    ""use strict"";

    function ns(selector, element) {
        return new NodeCollection(selector, element);
    }

    function NodeCollection(selector, element) {
        this.items = [];
        element = element || window.document;

        var nodeList;

        if (typeof (selector) === ""string"") {
            nodeList = element.querySelectorAll(selector);
            for (var i = 0, l = nodeList.length; i < l; i++) {
                this.items.push(nodeList.item(i));
            }
        }
    }

    NodeCollection.prototype = {
        each: function (callback) {
            for (var i = 0, l = this.items.length; i < l; i++) {
                callback(this.items[i], i);
            }
            return this;
        },

        children: function (selector) {
            var children = [];

            this.each(function (el) {
                children = children.concat(ns(selector, el)");
            WriteLiteral(@".items);
            });

            return ns(children);
        },

        hide: function () {
            this.each(function (el) {
                el.style.display = ""none"";
            });

            return this;
        },

        toggle: function () {
            this.each(function (el) {
                el.style.display = el.style.display === ""none"" ? """" : ""none"";
            });

            return this;
        },

        show: function () {
            this.each(function (el) {
                el.style.display = """";
            });

            return this;
        },

        addClass: function (className) {
            this.each(function (el) {
                var existingClassName = el.className,
                    classNames;
                if (!existingClassName) {
                    el.className = className;
                } else {
                    classNames = existingClassName.split("" "");
                    if (classNames.indexOf(className");
            WriteLiteral(@") < 0) {
                        el.className = existingClassName + "" "" + className;
                    }
                }
            });

            return this;
        },

        removeClass: function (className) {
            this.each(function (el) {
                var existingClassName = el.className,
                    classNames, index;
                if (existingClassName === className) {
                    el.className = """";
                } else if (existingClassName) {
                    classNames = existingClassName.split("" "");
                    index = classNames.indexOf(className);
                    if (index > 0) {
                        classNames.splice(index, 1);
                        el.className = classNames.join("" "");
                    }
                }
            });

            return this;
        },

        attr: function (name) {
            if (this.items.length === 0) {
                return null;
            }

         ");
            WriteLiteral(@"   return this.items[0].getAttribute(name);
        },

        on: function (eventName, handler) {
            this.each(function (el, idx) {
                var callback = function (e) {
                    e = e || window.event;
                    if (!e.which && e.keyCode) {
                        e.which = e.keyCode; // Normalize IE8 key events
                    }
                    handler.apply(el, [e]);
                };

                if (el.addEventListener) { // DOM Events
                    el.addEventListener(eventName, callback, false);
                } else if (el.attachEvent) { // IE8 events
                    el.attachEvent(""on"" + eventName, callback);
                } else {
                    el[""on"" + type] = callback;
                }
            });

            return this;
        },

        click: function (handler) {
            return this.on(""click"", handler);
        },

        keypress: function (handler) {
            return this.on(");
            WriteLiteral(@"""keypress"", handler);
        }
    };

    function frame(el) {
        ns("".source .collapsible"", el).toggle();
    }

    function expandCollapseButton(el) {
        var frameId = el.getAttribute(""data-frameId"");
        frame(document.getElementById(frameId));
        if (el.innerText === ""+"") {
            el.innerText = ""-"";
        }
        else {
            el.innerText = ""+"";
        }
    }

    function tab(el) {
        var unselected = ns(""#header .selected"").removeClass(""selected"").attr(""id"");
        var selected = ns(""#"" + el.id).addClass(""selected"").attr(""id"");

        ns(""#"" + unselected + ""page"").hide();
        ns(""#"" + selected + ""page"").show();
    }

    ns("".rawExceptionDetails"").hide();
    ns("".collapsible"").hide();
    ns("".page"").hide();
    ns(""#stackpage"").show();

    ns("".expandCollapseButton"")
        .click(function () {
            expandCollapseButton(this);
        })
        .keypress(function (e) {
            if (e.which === 13) {");
            WriteLiteral(@"
                expandCollapseButton(this);
            }
        });

    ns(""#header li"")
        .click(function () {
            tab(this);
        })
        .keypress(function (e) {
            if (e.which === 13) {
                tab(this);
            }
        });

    ns("".showRawException"")
        .click(function () {
            var exceptionDetailId = this.getAttribute(""data-exceptionDetailId"");
            ns(""#"" + exceptionDetailId).toggle();
        });
})(window);
            //-->
        </script>
</body>
</html>
");
        }
        #pragma warning restore 1998
#nullable restore
#line 9 "Views/ErrorPage.cshtml"
 
    public ErrorPage(ErrorPageModel model)
    {
        Model = model;
    }

    public ErrorPageModel Model { get; set; }

#line default
#line hidden
#nullable disable
    }
}
#pragma warning restore 1591
