namespace Microsoft.DotNet.SwaggerGenerator
{
    public class GeneratorOptions
    {
        public string Namespace { get; set; }
        public string ClientName { get; set; }
        public string LanguageName { get; set; }
    }
}
