// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

/**
 * This package contains the classes for SignalR Java client.
 */
package com.microsoft.signalr;
