// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

package com.microsoft.signalr;

/**
 *  Used to specify the transport the client will use.
 */
public enum TransportEnum {
    ALL,
    WEBSOCKETS,
    LONG_POLLING
}
